#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pytools',
 'pytools.api',
 'pytools.data',
 'pytools.expression',
 'pytools.expression.atomic',
 'pytools.expression.base',
 'pytools.expression.composite',
 'pytools.expression.composite.base',
 'pytools.expression.formatter',
 'pytools.expression.operator',
 'pytools.fit',
 'pytools.meta',
 'pytools.parallelization',
 'pytools.sphinx',
 'pytools.text',
 'pytools.viz',
 'pytools.viz.color',
 'pytools.viz.dendrogram',
 'pytools.viz.dendrogram.base',
 'pytools.viz.distribution',
 'pytools.viz.distribution.base',
 'pytools.viz.matrix',
 'pytools.viz.matrix.base',
 'pytools.viz.util']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['joblib         >=0.13,<1.1',
 'matplotlib     >=3.0,<3.4',
 'numpy          >=1.16,<1.21',
 'pandas         >=0.24,<1.3',
 'scipy          >=1.2,<1.6',
 'typing_inspect >=0.4,<0.7']

extras_require = \
{'docs': ['sphinx == 3.4.*',
          'sphinx-autodoc-typehints == 1.11.*',
          'pydata-sphinx-theme == 0.4.*',
          'nbsphinx == 0.7.*',
          'jupyter >= 1.0',
          'docutils',
          'xlrd == 1.2.*',
          'm2r == 0.2.*'],
 'testing': ['pytest >= 5.2.*',
             'pytest-cov == 2.8.*',
             'flake8 == 3.8.*',
             'flake8-comprehensions == 3.2.*',
             'isort == 5.5.*']}

setup(name='gamma-pytools',
      version='1.0.4',
      description="A collection of Python extensions and tools used in BCG GAMMA's open-source libraries.",
      author='Boston Consulting Group (BCG)',
      author_email=None,
      url='https://github.com/BCG-Gamma/pytools',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6,<4',
     )
